/*
 * Copyright (c) 2007 Bobby Bingham
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * audio and video splitter
 */

#include <stdio.h>

#include "libavutil/attributes.h"
#include "libavutil/avstring.h"
#include "libavutil/internal.h"
#include "libavutil/mem.h"
#include "libavutil/opt.h"
#include "libavutil/hwcontext_vpe.h"
#include "libavutil/hwcontext.h"

#include "avfilter.h"
#include "audio.h"
#include "filters.h"
#include "formats.h"
#include "internal.h"
#include "video.h"

#include "log_ama.h"
#include "vpi_buffer.h"

typedef struct SplitContext {
    const AVClass *class;
    int nb_outputs;
    char instance_name[128];
} SplitContext;

static av_cold int split_init(AVFilterContext *ctx)
{
    SplitContext *s = ctx->priv;
    int i, ret;

    for (i = 0; i < s->nb_outputs; i++) {
        AVFilterPad pad = { 0 };

        pad.type = ctx->filter->inputs[0].type;
        pad.name = av_asprintf("output%d", i);
        if (!pad.name)
            return AVERROR(ENOMEM);

        if ((ret = ff_append_outpad_free_name(ctx, &pad)) < 0)
            return ret;
    }

    snprintf(s->instance_name, 128, "%s@%p", (*(AVClass **)ctx)->item_name(ctx), ctx);

    return 0;
}

typedef struct SplitFrames {
    int count_freed;
    int count_alloc;
    VpiFrm **v_frames;
}SplitFrames;

static void av_buffer_default_free_vpi(void *opaque, uint8_t *data)
{
    SplitFrames *sf = opaque;
    if(!sf){
        av_log(NULL, AV_LOG_WARNING, "NULL opaque something is wrong !!\n");
        return;
    }
    sf->count_freed++;
    LOG_AMA_TRACE(LOG_AMA_LAYER_FFMPEG, "freeing split frame %d %d %p", sf->count_alloc, sf->count_freed, data);
    if(sf->count_alloc == sf->count_freed){
        LOG_AMA_TRACE(LOG_AMA_LAYER_FFMPEG, "split freeing everything !!");
        for(int i=0;i<sf->count_alloc;i++){
            if(sf->v_frames[i]){
                LOG_AMA_TRACE(LOG_AMA_LAYER_FFMPEG, "split freeing vpifrm %p !!", sf->v_frames[i]);
                av_free(sf->v_frames[i]);
            }
        }
        free(sf->v_frames);
        free(sf);
    }
}

static int _activate(AVFilterContext *ctx)
{
    AVFilterLink *inlink = ctx->inputs[0];
    AVFrame *in;
    int status, ret;
    int64_t pts;
    AVHWFramesContext *hwframe_ctx;
    SplitFrames *sf = NULL;

    for (int i = 0; i < ctx->nb_outputs; i++) {
        FF_FILTER_FORWARD_STATUS_BACK_ALL(ctx->outputs[i], ctx);
    }

    ret = ff_inlink_consume_frame(inlink, &in);
    if (ret < 0)
        return ret;

    if(ret > 0 && in->hw_frames_ctx){
        sf = (SplitFrames *)calloc(1, sizeof(SplitFrames));
        if(!sf){
            return AVERROR(ENOMEM);
        }
        sf->v_frames = (VpiFrm **)calloc(ctx->nb_outputs, sizeof(VpiFrm*));
        if(!sf->v_frames){
            free(sf);
            return AVERROR(ENOMEM);
        }
        sf->count_alloc = ctx->nb_outputs;
    }

    if (ret > 0) {
        for (int i = 0; i < ctx->nb_outputs; i++) {
            AVFrame *buf_out;

            if (ff_outlink_get_status(ctx->outputs[i]))
                continue;
            
            if(in->hw_frames_ctx) {
                VpiFrm *in_pic;
                buf_out = av_frame_alloc();
                if(!buf_out){
                    return AVERROR(ENOMEM);
                }
                av_frame_copy_props(buf_out, in);
                sf->v_frames[i] = av_malloc(sizeof(VpiFrm));
                if(!sf->v_frames[i]){
                    return AVERROR(ENOMEM);
                }
                memset(sf->v_frames[i], 0, sizeof(VpiFrm));
                LOG_AMA_TRACE(LOG_AMA_LAYER_FFMPEG, "Allocating vpi_frame(split) %p\n", sf->v_frames[i]);
                buf_out->buf[0] = av_buffer_create((uint8_t*)sf->v_frames[i], sizeof(VpiFrm), av_buffer_default_free_vpi, sf, 0);
                buf_out->data[0] = buf_out->buf[0]->data;

                if (in->hw_frames_ctx) {
                    buf_out->hw_frames_ctx = av_buffer_ref(in->hw_frames_ctx);
                    if (!buf_out->hw_frames_ctx) {
                        return AVERROR(ENOMEM);
                    }
                }

                hwframe_ctx = (AVHWFramesContext*)in->hw_frames_ctx->data;
                buf_out->format = hwframe_ctx->format;
                buf_out->width  = hwframe_ctx->width;
                buf_out->height = hwframe_ctx->height;

                in_pic = (VpiFrm *)(in->data[0]);

                memcpy(sf->v_frames[i], in_pic, sizeof(VpiFrm));
                memset(sf->v_frames[i]->vpi_sidedata, 0, sizeof(sf->v_frames[i]->vpi_sidedata));

                //HACK:: Because decoder's consume frames uses data inside VpiFrm
                if(((VpiBuffer*)(in_pic->vpi_picinfo_buf))->split_frame == NULL){
                    VpiFrm * opaque_frm = malloc(sizeof(VpiFrm));
                    memcpy(opaque_frm, in_pic, sizeof(VpiFrm));
                    ((VpiBuffer*)(in_pic->vpi_picinfo_buf))->split_frame = opaque_frm;
                    LOG_AMA_TRACE(LOG_AMA_LAYER_FFMPEG, "Setting data of %p to NULL\n", sf->v_frames[i]->vpi_picinfo_buf);
                    vpi_buffer_set_data(sf->v_frames[i]->vpi_picinfo_buf, NULL);

                }
                //update sidedata
                vpifrm_sidedata_copy_reference(in_pic, sf->v_frames[i]);
                vpi_frame_imagedata_ref(sf->v_frames[i]);

            } else {
                buf_out = av_frame_clone(in);
                if (!buf_out) {
                    ret = AVERROR(ENOMEM);
                    break;
                }
                if (i && buf_out->hw_frames_ctx) {
                hwframe_ctx = (AVHWFramesContext*)buf_out->hw_frames_ctx->data;
                if (hwframe_ctx) {
                    if (hwframe_ctx->format == AV_PIX_FMT_VPE) {
                        if (buf_out->buf[0]) {
                            VpiFrm *vpi_frame = (VpiFrm*)buf_out->buf[0]->data;
                            if (vpi_frame) {
                                vpi_frame_ref(vpi_frame);
                                }
                            }
                        }
                    }
                }
            }

            ret = ff_filter_frame(ctx->outputs[i], buf_out);
            if (ret < 0)
                break;
        }

        if(in && in->data[0] && in->format == AV_PIX_FMT_VPE){
            vpi_frame_unref((VpiFrm*)in->data[0]);
        }
        av_frame_free(&in);
        if (ret < 0)
            return ret;
    }

    if (ff_inlink_acknowledge_status(inlink, &status, &pts)) {
        for (int i = 0; i < ctx->nb_outputs; i++) {
            if (ff_outlink_get_status(ctx->outputs[i]))
                continue;
            ff_outlink_set_status(ctx->outputs[i], status, pts);
        }
        return 0;
    }

    for (int i = 0; i < ctx->nb_outputs; i++) {
        if (ff_outlink_get_status(ctx->outputs[i]))
            continue;

        if (ff_outlink_frame_wanted(ctx->outputs[i])) {
            ff_inlink_request_frame(inlink);
            return 0;
        }
    }

    return FFERROR_NOT_READY;
}

static int activate(AVFilterContext *ctx)
{
    int ret;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, ((SplitContext*)(ctx->priv))->instance_name);
    ret = _activate(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}

#define OFFSET(x) offsetof(SplitContext, x)
#define FLAGS (AV_OPT_FLAG_AUDIO_PARAM | AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM)
static const AVOption options[] = {
    { "outputs", "set number of outputs", OFFSET(nb_outputs), AV_OPT_TYPE_INT, { .i64 = 2 }, 1, INT_MAX, FLAGS },
    { NULL }
};

AVFILTER_DEFINE_CLASS_EXT(split, "(a)split", options);

static const AVFilterPad avfilter_vf_split_inputs[] = {
    {
        .name         = "default",
        .type         = AVMEDIA_TYPE_VIDEO,
    },
};

const AVFilter ff_vf_split = {
    .name        = "split",
    .description = NULL_IF_CONFIG_SMALL("Pass on the input to N video outputs."),
    .priv_size   = sizeof(SplitContext),
    .priv_class  = &split_class,
    .init        = split_init,
    .activate    = activate,
    FILTER_INPUTS(avfilter_vf_split_inputs),
    .outputs     = NULL,
    .flags       = AVFILTER_FLAG_DYNAMIC_OUTPUTS | AVFILTER_FLAG_METADATA_ONLY,
};

static const AVFilterPad avfilter_af_asplit_inputs[] = {
    {
        .name         = "default",
        .type         = AVMEDIA_TYPE_AUDIO,
    },
};

const AVFilter ff_af_asplit = {
    .name        = "asplit",
    .description = NULL_IF_CONFIG_SMALL("Pass on the audio input to N audio outputs."),
    .priv_class  = &split_class,
    .priv_size   = sizeof(SplitContext),
    .init        = split_init,
    .activate    = activate,
    FILTER_INPUTS(avfilter_af_asplit_inputs),
    .outputs     = NULL,
    .flags       = AVFILTER_FLAG_DYNAMIC_OUTPUTS | AVFILTER_FLAG_METADATA_ONLY,
};
