/*
 * VeriSilicon VPE Post Processing Filter
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERC`ABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdint.h>

#include "vpi_types.h"
#include "vpi_api.h"

#include "filters.h"
#include "libavutil/opt.h"
#include "libavutil/hwcontext.h"
#include "libavutil/hwcontext_vpe.h"
#include "libavutil/pixdesc.h"

typedef struct VpePPFilterContext {
    const AVClass *av_class;
    AVBufferRef *hw_device;
    AVBufferRef *hw_frame;

    VpiProcInitParams *init_params;
    VpiProcHandle vpe_pp_ctx;

    /*pp option*/
    int nb_outputs;
    int rgb_enable;
    char *low_res;
    char *rgb_fmt;
    int force8;
    char *rgb_size;
    char *extra_params;

    int max_lowres;
    int planar;
    AVDictionary *dict;
} VpePPFilterContext;

static int vpe_pp_VCDA_preinit(AVFilterContext *ctx)
{
    VpePPFilterContext *s = ctx->priv;
    s->init_params->processor_type = VPE_PROCESSOR_PP_VCDA;
    s->max_lowres = 6;
    return 0;
}

static int vpe_pp_VCDB_preinit(AVFilterContext *ctx)
{
    VpePPFilterContext *s = ctx->priv;
    s->init_params->processor_type = VPE_PROCESSOR_PP_VCDB;
    s->max_lowres = 2;
    return 0;
}

static av_cold int vpe_pp_init(AVFilterContext *ctx)
{
    VpePPFilterContext *s = ctx->priv;
    AVDictionaryEntry *en = NULL;
    AVFilterPad pad = { 0 };
    int i = 0;
    int ret = 0;

    av_log(ctx, AV_LOG_DEBUG, "ctx->hw_device_ctx = %p\n", ctx->hw_device_ctx);
    s->init_params =
        (VpiProcInitParams *)vpi_memory_get_struct(VPE_PROC_INIT_PARAM);
    if (!s->init_params) {
        av_log(ctx, AV_LOG_DEBUG, "pp alloc init param failed\n");
        return AVERROR(ENOMEM);
    }

    /* process options and put options into dict */
    if (s->extra_params) {
        if ((ret = av_dict_parse_string(&s->dict, s->extra_params, "=", "|", 0)) < 0) {
            return ret;
        }
        while ((en = av_dict_get(s->dict, "", en, AV_DICT_IGNORE_SUFFIX))) {
            av_log(ctx, AV_LOG_DEBUG, "key = %s, value = %s\n", en->key, en->value);
        }
        en = NULL;
        if ((en = av_dict_get(s->dict, "planar", en, AV_DICT_IGNORE_SUFFIX))) {
            if (!strcmp(en->value, "1")) {
                s->planar = 1;
            }
        }
    }

    if (s->rgb_enable) {
        if (!s->rgb_size) {
            return AVERROR_INVALIDDATA;
        }
        if ((ret = av_dict_set(&s->dict, "rgb_enable", "1", 0)) < 0) {
            return ret;
        }
        if ((ret = av_dict_set(&s->dict, "rgb_size", s->rgb_size, 0)) < 0) {
            return ret;
        }
        if (s->rgb_fmt) {
            if ((ret = av_dict_set(&s->dict, "rgb_fmt", s->rgb_fmt, 0)) < 0) {
                return ret;
            }
        }
    }

    if (s->force8) {
        if ((ret = av_dict_set(&s->dict, "force8", "1", 0)) < 0) {
            return ret;
        }
        en = NULL;
        if ((en = av_dict_get(s->dict, "force8", en, AV_DICT_IGNORE_SUFFIX))) {
            av_log(ctx, AV_LOG_DEBUG, "key = %s, value = %s\n", en->key, en->value);
        }
    }

    pad.type = AVMEDIA_TYPE_VIDEO;
    pad.name = "output0";
    if ((ret = ff_insert_outpad(ctx, i, &pad)) < 0) {
        return ret;
    }

    return 0;
}

static av_cold void vpe_pp_uninit(AVFilterContext *ctx)
{
    VpePPFilterContext *s = ctx->priv;

    if (s->vpe_pp_ctx != NULL) {
        vpi_processor_close(s->vpe_pp_ctx);
    }

    if (s->init_params) {
        vpi_memory_free_struct(s->init_params);
    }
    av_dict_free(&s->dict);
    av_buffer_unref(&s->hw_frame);
    av_buffer_unref(&s->hw_device);
}

static int vpe_pp_filter_frame(AVFilterLink *inlink, AVFrame *frame)
{
    AVFilterContext *ctx  = inlink->dst;
    VpePPFilterContext *s = ctx->priv;
    AVFilterLink *outlink = ctx->outputs[0];
    AVFrame *pp_frame;
    AVHWFramesContext *hwframe_ctx;
    VpiFrm *in_picture, *out_picture;
    int ret = 0;

    hwframe_ctx  = (AVHWFramesContext *)s->hw_frame->data;

    if (inlink->format != AV_PIX_FMT_VPE) {
        return AVERROR_UNKNOWN;
    } else {
        in_picture = (VpiFrm *)frame->data[0];
    }

    pp_frame = ff_get_video_buffer(outlink, outlink->w, outlink->h);
    if (!pp_frame) {
        av_log(outlink, AV_LOG_ERROR, "Failed to allocate frame to pp_vpe\n");
        ret = AVERROR(ENOMEM);
        goto fail;
    }

    out_picture = (VpiFrm *)pp_frame->data[0];
    ret = vpi_processor_process_frame(s->vpe_pp_ctx, in_picture, out_picture);
    if (ret) {
        av_log(NULL, AV_LOG_ERROR, "VPE PP error: %s\n", vpi_error_str(ret));
        ret = AVERROR_EXTERNAL;
        goto fail;
    }

    ret = av_frame_copy_props(pp_frame, frame);
    if (ret)
        goto fail;

    pp_frame->width       = out_picture->width;
    pp_frame->height      = out_picture->height;
    pp_frame->key_frame   = frame->key_frame;
    pp_frame->format      = hwframe_ctx->sw_format;

    ret = ff_outlink_get_status(outlink);
    if (ret < 0)
        goto fail;

    av_frame_free(&frame);
    return ff_filter_frame(outlink, pp_frame);

fail:
    av_frame_free(&frame);
    av_frame_free(&pp_frame);
    return ret;
}

static int vpe_pp_check_constraints(AVFilterLink *inlink)
{
    AVFilterContext *ctx = inlink->dst;
    VpePPFilterContext *s = ctx->priv;
    AVVpeHWConfig *hwconfig = NULL;
    AVHWFramesConstraints *constraints = NULL;
    int ret = 0;

    hwconfig = av_hwdevice_hwconfig_alloc(s->hw_device);
    if (!hwconfig) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    hwconfig->id = VPE_HW_PP;
    av_log(ctx, AV_LOG_DEBUG, "hwconfig->id = %d\n", hwconfig->id);

    constraints =
        av_hwdevice_get_hwframe_constraints(s->hw_device, hwconfig);
    if (!constraints) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    av_log(ctx, AV_LOG_DEBUG, "get constraints min_w %d, min_h %d, "
                              "max_w %d, max_h %d\n",
                               constraints->min_width, constraints->min_height,
                               constraints->max_width, constraints->max_height);

    if (inlink->w < constraints->min_width  ||
        inlink->h < constraints->min_height ||
        inlink->w > constraints->max_width  ||
        inlink->h > constraints->max_height) {
        av_log(ctx, AV_LOG_ERROR, "Hardware does not support image "
               "size %dx%d (constraints: width %d-%d height %d-%d).\n",
               inlink->w, inlink->h,
               constraints->min_width,  constraints->max_width,
               constraints->min_height, constraints->max_height);
        ret = AVERROR(EINVAL);
    }

    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);

    return ret;

error:
    av_log(ctx, AV_LOG_ERROR, "vpe_pp_check_constraints failed\n");

    if (constraints) {
        av_hwframe_constraints_free(&constraints);
    }
    if (hwconfig) {
        av_freep(&hwconfig);
    }

    return ret;

}

static int vpe_pp_init_hwctx(AVFilterLink *inlink)
{
    AVFilterContext *ctx = inlink->dst;
    VpePPFilterContext *s = ctx->priv;
    VpiProcInitParams *param = s->init_params;
    AVHWFramesContext *hwframe_ctx;
    AVVpeFramesContext *frame_hwctx;
    AVHWDeviceContext *device_ctx;
    AVVpeDeviceContext *device_hwctx;
    int ret;

    /* hw device & frame init */
    if (inlink->hw_frames_ctx) {
        s->hw_frame = av_buffer_ref(inlink->hw_frames_ctx);
        if (!s->hw_frame) {
            return AVERROR(ENOMEM);
        }

        hwframe_ctx = (AVHWFramesContext*)s->hw_frame->data;
    } else {
        av_log(ctx, AV_LOG_ERROR, "vpe_pp can't get hwframe\n");
        return AVERROR_UNKNOWN;
    }

    ret = vpe_pp_check_constraints(inlink);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpe_pp_check_constraints failed\n");
        return ret;
    }

    device_ctx   = hwframe_ctx->device_ctx;
    device_hwctx = device_ctx->hwctx;
    vpi_device_set_ctx(param, device_hwctx->vpe_dev_ctx, VPE_HW_PP);

    frame_hwctx = hwframe_ctx->hwctx;
    param->mem_ctx = frame_hwctx->vpe_mem_ctx;

    if (s->rgb_enable && s->nb_outputs == 0) {
        if (!s->rgb_fmt || strcmp(s->rgb_fmt, "rgb") == 0) {
            if (s->planar) {
                if (hwframe_ctx->sw_format == AV_PIX_FMT_NV12) {
                    frame_hwctx->sw_format[0] = AV_PIX_FMT_RGB24;
                } else if (hwframe_ctx->sw_format == AV_PIX_FMT_P010LE) {
                    frame_hwctx->sw_format[0] = AV_PIX_FMT_RGB24;
                } else {
                    return AVERROR_INVALIDDATA;
                }
            } else {
                if (hwframe_ctx->sw_format == AV_PIX_FMT_NV12) {
                    frame_hwctx->sw_format[0] = AV_PIX_FMT_RGB24;
                } else if (hwframe_ctx->sw_format == AV_PIX_FMT_P010LE) {
                    frame_hwctx->sw_format[0] = AV_PIX_FMT_RGB24;
                } else {
                    return AVERROR_INVALIDDATA;
                }
            }
        } else if (strcmp(s->rgb_fmt, "bgr") == 0) {
            if (s->planar) {
                if (hwframe_ctx->sw_format == AV_PIX_FMT_NV12) {
                    frame_hwctx->sw_format[0] = AV_PIX_FMT_RGB24;
                } else if (hwframe_ctx->sw_format == AV_PIX_FMT_P010LE) {
                    frame_hwctx->sw_format[0] = AV_PIX_FMT_RGB24;
                } else {
                    return AVERROR_INVALIDDATA;
                }
            } else {
                if (hwframe_ctx->sw_format == AV_PIX_FMT_NV12) {
                    frame_hwctx->sw_format[0] = AV_PIX_FMT_BGR24;
                } else if (hwframe_ctx->sw_format == AV_PIX_FMT_P010LE) {
                    frame_hwctx->sw_format[0] = AV_PIX_FMT_RGB24;
                } else {
                    return AVERROR_INVALIDDATA;
                }
            }
        } else {
            frame_hwctx->sw_format[0] = av_get_pix_fmt(s->rgb_fmt);
            if (frame_hwctx->sw_format[0] == AV_PIX_FMT_NONE) {
                return AVERROR_OPTION_NOT_FOUND;
            }
        }
    } else {
        frame_hwctx->sw_format[0] = hwframe_ctx->sw_format; //default equal to input, and need add adjust depend on option
        if (s->planar) {
            if (hwframe_ctx->sw_format == AV_PIX_FMT_NV12) {
                frame_hwctx->sw_format[0] = AV_PIX_FMT_YUV420P;
            } else if (hwframe_ctx->sw_format == AV_PIX_FMT_P010LE) {
                frame_hwctx->sw_format[0] = AV_PIX_FMT_YUV420P10LE;
            } else {
                return AVERROR_INVALIDDATA;
            }
        }
        if (s->force8) {
            if (frame_hwctx->sw_format[0] == AV_PIX_FMT_YUV420P10LE) {
                frame_hwctx->sw_format[0] = AV_PIX_FMT_YUV420P;
            }
            if (frame_hwctx->sw_format[0] == AV_PIX_FMT_P010LE) {
                frame_hwctx->sw_format[0] = AV_PIX_FMT_NV12;
            }
        }
    }

    return 0;
}

static int vpe_pp_config_props(AVFilterLink *inlink)
{
    AVFilterContext *ctx = inlink->dst;
    VpePPFilterContext *s = ctx->priv;
    VpiProcInitParams *params = s->init_params;
    AVHWFramesContext *hwframe_ctx;
    AVVpeFramesContext *frame_hwctx;
    AVDictionaryEntry *en = NULL;
    int i;
    int ret;
    int out_en_count;

    ret = vpe_pp_init_hwctx(inlink);
    if (ret < 0) {
        return ret;
    }

    hwframe_ctx = (AVHWFramesContext *)s->hw_frame->data;
    if (!hwframe_ctx) {
        av_log(ctx, AV_LOG_ERROR, "hwframe_ctx is nil\n");
        return AVERROR_UNKNOWN;
    }

    frame_hwctx = hwframe_ctx->hwctx;
    if (!frame_hwctx) {
        av_log(ctx, AV_LOG_ERROR, "frame_hwctx is nil\n");
        return AVERROR_UNKNOWN;
    }

    params->output_num     = s->nb_outputs;
    params->width          = inlink->w;
    params->height         = inlink->h;
    params->input_format   = av_get_pix_fmt_name(inlink->format);
    if (inlink->format == AV_PIX_FMT_VPE) {
        AVHWFramesContext *fmctx = (AVHWFramesContext *)inlink->hw_frames_ctx->data;
        params->input_sw_format = av_get_pix_fmt_name(fmctx->sw_format);
    } else {
        params->input_sw_format = av_get_pix_fmt_name(inlink->format);
    }
    params->resize_str     = s->low_res;

    ret = vpi_processor_init(params, &s->vpe_pp_ctx);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "VPE processor vf_pp init error: %s\n", vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    while ((en = av_dict_get(s->dict, "", en, AV_DICT_IGNORE_SUFFIX))) {
        if (vpi_processor_set_param(s->vpe_pp_ctx, en->key, en->value) < 0)
            av_log(ctx, AV_LOG_WARNING,
                   "Error parsing option '%s = %s'. Please check if it is valid for vpe pp options.\n",
                   en->key, en->value);
    }

    ret = vpi_processor_open(s->vpe_pp_ctx);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_processor_open failed, %s\n", vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    ret = vpi_processor_get_info(s->vpe_pp_ctx, frame_hwctx->proc_info);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_processor_get_info failed\n");
        return AVERROR_EXTERNAL;
    }

    out_en_count = 0;
    for (i = 0; i < VPE_CHANNEL_NUM; i++) {
        if (frame_hwctx->proc_info->ch_info[i].enabled) {
            out_en_count++;
            av_log(ctx, AV_LOG_DEBUG, "info.ch_info[%d].width = %d\n",
                                       i, frame_hwctx->proc_info->ch_info[i].width);
            av_log(ctx, AV_LOG_DEBUG, "info.ch_info[%d].height = %d\n",
                                       i, frame_hwctx->proc_info->ch_info[i].height);
            av_log(ctx, AV_LOG_DEBUG, "info.ch_info[%d].format = %d\n",
                                       i, frame_hwctx->proc_info->ch_info[i].format);
            av_log(ctx, AV_LOG_DEBUG, "info.ch_info[%d].sw_format = %d\n",
                                       i, frame_hwctx->proc_info->ch_info[i].sw_format);
        }
    }

    if (out_en_count == 0) {
        av_log(ctx, AV_LOG_ERROR, "output channels are all invalid\n");
        return AVERROR_EXTERNAL;
    }

    return 0;
}

static int vpe_pp_query_formats(AVFilterContext *ctx)
{
    VpePPFilterContext *s = ctx->priv;
    AVVpeHWConfig *hwconfig = NULL;
    AVHWFramesConstraints *constraints = NULL;
    AVFilterFormats *in_fmts;
    AVFilterFormats *out_fmts;
    int ret;
    int i;

    av_log(ctx, AV_LOG_DEBUG, "ctx->hw_device_ctx = %p\n", ctx->hw_device_ctx);

    if (!s->hw_device) {
        if (ctx->hw_device_ctx) {
            s->hw_device = av_buffer_ref(ctx->hw_device_ctx);
            if (!s->hw_device) {
                return AVERROR(ENOMEM);
            }
        } else {
            return AVERROR_UNKNOWN;
        }
    }

    hwconfig = av_hwdevice_hwconfig_alloc(s->hw_device);
    if (!hwconfig) {
        return AVERROR(ENOMEM);
    }

    hwconfig->id = VPE_HW_PP;

    constraints =
        av_hwdevice_get_hwframe_constraints(s->hw_device, hwconfig);
    if (!constraints) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    out_fmts = ff_make_format_list(constraints->valid_hw_formats);
    av_log(ctx, AV_LOG_DEBUG, "out_fmts->nb_formats = %d\n",
                               out_fmts->nb_formats);
    ret = ff_formats_ref(out_fmts, &ctx->outputs[0]->incfg.formats);
    if (ret < 0)
        goto error;

    in_fmts = ff_make_format_list(constraints->valid_hw_formats);
    i = 0;
    while (constraints->valid_sw_formats[i] != AV_PIX_FMT_NONE) {
        av_log(ctx, AV_LOG_DEBUG, "constraints->valid_sw_formats[%d] = %s\n",
               i, av_get_pix_fmt_name(constraints->valid_sw_formats[i]));
        ret = ff_add_format(&in_fmts, constraints->valid_sw_formats[i]);
        if (ret < 0)
            goto error;
        i++;
    }
    ret = ff_formats_ref(in_fmts, &ctx->inputs[0]->outcfg.formats);
    if (ret < 0)
        goto error;

    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);

    return 0;

error:
    av_log(ctx, AV_LOG_ERROR, "vpe_pp_query_formats failed\n");

    if (constraints) {
        av_hwframe_constraints_free(&constraints);
    }
    if (hwconfig) {
        av_freep(&hwconfig);
    }

    return ret;
}

#define OFFSET(x) offsetof(VpePPFilterContext, x)
#define FLAGS (AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM)

static const AVOption vpe_pp_options[] = {
    { "outputs",
      "set number of outputs",
      OFFSET(nb_outputs),
      AV_OPT_TYPE_INT,
      { .i64 = 1 },
      1,
      6,
      FLAGS },
    { "low_res",
      "specific resize configuration.",
      OFFSET(low_res),
      AV_OPT_TYPE_STRING,
      { .str = NULL },
      .flags = FLAGS },
    { "rgb_enable",
      "enable output rgb data",
      OFFSET(rgb_enable),
      AV_OPT_TYPE_INT,
      { .i64 = 0 },
      0,
      1,
      FLAGS },
    { "rgb_fmt",
      "set rgb output format",
      OFFSET(rgb_fmt),
      AV_OPT_TYPE_STRING,
      { .str = NULL },
      .flags = FLAGS },
    { "force8",
      "enable output cut to 8bit",
      OFFSET(force8),
      AV_OPT_TYPE_INT,
      { .i64 = 0 },
      0,
      1,
      .flags = FLAGS },
    { "rgb_size",
      "set rgb output downscale size",
      OFFSET(rgb_size),
      AV_OPT_TYPE_STRING,
      { .str = NULL },
      .flags = FLAGS },
    { "pp_params",
      "set pp extra params",
      OFFSET(extra_params),
      AV_OPT_TYPE_STRING,
      { .str = NULL },
      .flags = FLAGS },
    { NULL },
};

AVFILTER_DEFINE_CLASS(vpe_pp);

static const AVFilterPad vpe_pp_inputs[] = {
    {
        .name         = "default",
        .type         = AVMEDIA_TYPE_VIDEO,
        .filter_frame = vpe_pp_filter_frame,
        .config_props = vpe_pp_config_props,
    },
    { NULL }
};

AVFilter ff_vf_pp_vpe_VCDA = {
    .name          = "pp_vpeA",
    .description   = NULL_IF_CONFIG_SMALL("Filter using vpe post processing of VC8000D_A."),
    .priv_size     = sizeof(VpePPFilterContext),
    .priv_class    = &vpe_pp_class,
    .preinit       = vpe_pp_VCDA_preinit,
    .init          = vpe_pp_init,
    .uninit        = vpe_pp_uninit,
    .query_formats = vpe_pp_query_formats,
    .inputs        = vpe_pp_inputs,
    .outputs       = NULL,
    .flags         = 0,
};

AVFilter ff_vf_pp_vpe_VCDB = {
    .name          = "pp_vpeB",
    .description   = NULL_IF_CONFIG_SMALL("Filter using vpe post processing of VC8000D_B."),
    .priv_size     = sizeof(VpePPFilterContext),
    .priv_class    = &vpe_pp_class,
    .preinit       = vpe_pp_VCDB_preinit,
    .init          = vpe_pp_init,
    .uninit        = vpe_pp_uninit,
    .query_formats = vpe_pp_query_formats,
    .inputs        = vpe_pp_inputs,
    .outputs       = NULL,
    .flags         = 0,
};
