/*
* AMD ROI side data scale filter
*
* This file is part of FFmpeg.
*
* FFmpeg is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* FFmpeg is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "roi_scale.h"
#include "vpi_types.h"
#include "vpi_api.h"

#include "avfilter.h"
#include "filters.h"
#include "framesync.h"
#include "libavutil/avstring.h"
#include "libavutil/opt.h"
#include "libavutil/pixdesc.h"
#include "log_ama.h"

#define ROI_SCALE_MAX_INPUTS 4

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

typedef enum {
    ML_ROI_DQP = 0,
    ML_ROI_RAW,
    ML_ROI_DQP_AND_RAW
} MLROIType;

typedef struct ROIScaleContext {
    const AVClass *class;
    FFFrameSync fs;
    int nb_inputs;
    char *roi_map_type;
    MLROIType roi_map_val;
    int cutoff_per;
    char *static_map_file;
    unsigned char *src_roi;
    unsigned char *dst_roi;
    unsigned char * static_roi_map;
    char *strength;
    int strength_val;
    unsigned int src_roi_width;
    unsigned int src_roi_height;
    unsigned int src_roi_size;
    unsigned int dst_roi_width;
    unsigned int dst_roi_height;
    unsigned int dst_roi_size;
    VpiSidedataPoolManagerHandle sd_pool_mgr;
    char instance_name[64];
} ROIScaleContext;

static int roi_scale_query_formats(AVFilterContext *ctx)
{
    AVFilterFormats *formats = NULL;
    int fmt, ret;

    for (fmt = 0; av_pix_fmt_desc_get(fmt); fmt++) {
        const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get(fmt);
        if (desc->flags & AV_PIX_FMT_FLAG_BITSTREAM)
            continue;
        if (!(desc->flags & AV_PIX_FMT_FLAG_HWACCEL)) {
            // Not usable if there is any subsampling but the format is
            // not planar (e.g. YUYV422).
            if ((desc->log2_chroma_w || desc->log2_chroma_h) &&
                !(desc->flags & AV_PIX_FMT_FLAG_PLANAR))
                continue;
        }
        ret = ff_add_format(&formats, fmt);
        if (ret < 0)
            return ret;
    }

    return ff_set_common_formats(ctx, formats);
}

static int query_formats(AVFilterContext *ctx)
{
    int ret;
    ROIScaleContext *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, s->instance_name);
    ret = roi_scale_query_formats(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ret;
}


static av_cold int roi_scale_init(AVFilterContext *ctx)
{
    int ret;
    ROIScaleContext *s = NULL;

    s = ctx->priv;

    for (int i = 0; i < s->nb_inputs; i++) {
        AVFilterPad pad = { 0 };

        pad.type = AVMEDIA_TYPE_VIDEO;
        pad.name = av_asprintf("input%d", i);
        if (!pad.name)
            return AVERROR(ENOMEM);

        if ((ret = ff_append_inpad(ctx, &pad)) < 0) {
            av_freep(&pad.name);
            return ret;
        }
    }
    ret = vpi_sidedata_pool_manager_init(NULL, &s->sd_pool_mgr);
    if (ret) {
        av_log(ctx, AV_LOG_ERROR, "failed to create sd pool manger %d.\n", ret);
        return AVERROR(ret);
    }

    s->static_roi_map = NULL;
    s->src_roi_width = 0;
    s->src_roi_height = 0;
    s->src_roi_size = 0;
    s->dst_roi_width = 0;
    s->dst_roi_height = 0;
    s->dst_roi_size = 0;


    if (!av_strcasecmp("dqp", s->roi_map_type)) {
        s->roi_map_val = ML_ROI_DQP;
    } else if (!av_strcasecmp("roi", s->roi_map_type)) {
        s->roi_map_val = ML_ROI_RAW;
    } else if (!av_strcasecmp("dqp_and_roi", s->roi_map_type)) {
        s->roi_map_val = ML_ROI_DQP_AND_RAW;
    } else {            
        av_log(ctx, AV_LOG_ERROR, "Invalid roi_map_type value %s.\n", s->roi_map_type);
        return AVERROR(EINVAL);
    }

    if (!av_strcasecmp("Low", s->strength)) {
        s->strength_val = 4;
    } else if (!av_strcasecmp("Normal", s->strength)) {
        s->strength_val = 8;
    } else if (!av_strcasecmp("High", s->strength)) {
        s->strength_val = 16;
    } else {            
        av_log(ctx, AV_LOG_ERROR, "Invalid strength value %s.\n", s->strength);
        return AVERROR(EINVAL);
    }

    return 0;
}

static void set_instance_name(ROIScaleContext *s, const char* ffmpeg_name)
{
    char *ptr = strrchr(ffmpeg_name, '_');
    snprintf(s->instance_name, sizeof(s->instance_name), (ptr) ? "roi_scale_ama%s" : "roi_scale_ama@%p", (ptr) ? ptr : s);
}

static av_cold int init(AVFilterContext *ctx)
{
    int ret;
    ROIScaleContext *s = ctx->priv;
    set_instance_name(s, ctx->name);

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, s->instance_name);
    ret = roi_scale_init(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ret;
}


static av_cold void roi_scale_uninit(AVFilterContext *ctx)
{
    ROIScaleContext *s = NULL;

    s = ctx->priv;
    ff_framesync_uninit(&s->fs);

    if (s->sd_pool_mgr)
        vpi_sidedata_pool_manager_uninit(s->sd_pool_mgr);

    if (s->src_roi)
        free(s->src_roi);

    if (s->dst_roi)
        free(s->dst_roi);

    if (s->static_roi_map)
        free(s->static_roi_map);

    for (int i = 0; i < ctx->nb_inputs; i++)
        av_freep(&ctx->input_pads[i].name);
}

static av_cold void uninit(AVFilterContext *ctx)
{
    ROIScaleContext *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, s->instance_name);
    roi_scale_uninit(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();
}


static void sb_scale_attribution(VpiSidedataHandle orgsb, VpiSidedataHandle newsb, int width, int height)
{
    int nchw[4];
    int type = vpi_get_sidedata_type(newsb);
    int layout = vpi_get_sidedata_layout(orgsb);
    int data_format = vpi_get_sidedata_dataformat(orgsb);
    int batchsize = vpi_get_sidedata_batchsize(orgsb);
    int ch = vpi_get_sidedata_layoutch(orgsb);

    switch (layout) {
        case VPE_LAYOUT_NCHW:
            nchw[0] = width;
            nchw[1] = height;
            nchw[2] = ch;
            nchw[3] = batchsize;
            break;
        case VPE_LAYOUT_NHWC:
            nchw[0] = ch;
            nchw[1] = width;
            nchw[2] = height;
            nchw[3] = batchsize;
            break;
        case VPE_LAYOUT_CHWN:
            nchw[0] = batchsize;
            nchw[1] = width;
            nchw[2] = height;
            nchw[3] = ch;
            break;
        default:
            nchw[0] = width;
            nchw[1] = height;
            nchw[2] = ch;
            nchw[3] = batchsize;
            break;
    }
    vpi_update_side_data_attribute(newsb, type, nchw, layout, data_format);
    return;
}

static int process_frame(FFFrameSync *fs)
{
    AVFilterContext *ctx = NULL;
    ROIScaleContext *s = NULL;
    AVFilterLink *outlink = NULL;
    AVFrame *out, *in[ROI_SCALE_MAX_INPUTS];
    VpiFrm *roi_pic = NULL;
    VpiFrm *in_pic = NULL;
    VpiFrm *out_pic = NULL;
    VpiSidedataHandle src_roi_sd = NULL;
    int ret = 0;

    ctx = fs->parent;
    s = fs->opaque;
    outlink = ctx->outputs[0];

    for (int i = 0; i < s->nb_inputs; i++)
        ff_framesync_get_frame(&s->fs, i, &in[i], 0);

    out = ff_get_video_buffer(outlink, in[0]->width, in[0]->height);
    if (!out) {
        av_log(NULL, AV_LOG_ERROR, "failed to allocate frame.");
        return AVERROR(ENOMEM);
    }

    ret = av_frame_copy_props(out, in[0]);
    if (ret < 0){
        av_log(NULL, AV_LOG_ERROR, "av_frame_copy_props failed\n");
        return AVERROR_EXTERNAL;
    }

    out_pic = (VpiFrm *)out->data[0];
    if (!out_pic){
        return AVERROR_INVALIDDATA;
    }

    in_pic = (VpiFrm *)(in[0]->data[0]);
    out_pic->vpi_picinfo_buf      = in_pic->vpi_picinfo_buf;
    out_pic->width                = in_pic->width;
    out_pic->height               = in_pic->height;
    out_pic->format               = in_pic->format;
    out_pic->ch_index             = in_pic->ch_index;
    out_pic->log_frm_num          = in_pic->log_frm_num;
    out_pic->ch_info[out_pic->ch_index]    = in_pic->ch_info[in_pic->ch_index];

    vpifrm_sidedata_copy_reference(in_pic, out_pic);
    vpi_frame_imagedata_ref(out_pic);
    vpi_frame_unref(in_pic);

    bool roi_present = false;
    for (int i = 1; i < s->nb_inputs; i++) {
        roi_pic = (VpiFrm *)in[i]->data[0];
        vpifrm_get_first_sidedata_by_type(roi_pic, VPI_SB_ML_ROI, &src_roi_sd);
        while (src_roi_sd) {
            unsigned int width = vpi_get_sidedata_width(src_roi_sd);
            unsigned int height = vpi_get_sidedata_height(src_roi_sd);
            unsigned int size = vpi_get_sidedata_size(src_roi_sd);
            if ((width == s->src_roi_width) && (height == s->src_roi_height) && (size == s->src_roi_size)) {
                if (!roi_present) {
                    memcpy(s->src_roi, vpi_get_sidedata_vaddr(src_roi_sd), size);
                    roi_present = true;
                } else {
                    unsigned char *additional_roi = vpi_get_sidedata_vaddr(src_roi_sd);

                    for (int j = 0; j < size; j++) {
                        s->src_roi[j] = max(s->src_roi[j], additional_roi[j]);
                    }
                }
            } else {
                av_log(ctx, AV_LOG_ERROR, "ROI input-%d dimension is not as expected.\n", i);
                return AVERROR(EINVAL);
            }
            src_roi_sd = vpifrm_find_next_sidedata((VpiFrm *)roi_pic, src_roi_sd);
        }
    }

    if (s->static_roi_map) {
        for (int i=0; i<s->src_roi_size; ++i) {
            s->src_roi[i] = max(s->src_roi[i], s->static_roi_map[i]);
        }
    }

    roi_pic = (VpiFrm *)in[1]->data[0];
    vpifrm_get_first_sidedata_by_type(roi_pic, VPI_SB_ML_ROI, &src_roi_sd);

    if ((in[1]->width == out->width) && (in[1]->height == out->height)) {
        memcpy(s->dst_roi, s->src_roi, s->src_roi_size);
    } else {
        scale_roi((uint8_t*) s->src_roi, (uint8_t*) s->dst_roi, in[1]->width, in[1]->height, out->width, out->height);
    }

    int total_activations_cnt = 0;
    int total_activations_per = 0;
    for (int i=0; i<s->dst_roi_size; i++) {
        if (s->dst_roi[i] > 128) {
            ++total_activations_cnt;
        }
    }
    total_activations_per = (total_activations_cnt*100000)/s->dst_roi_size;

    if (total_activations_per > s->cutoff_per) {
        for (int i=0; i<s->dst_roi_size; i++) {
            s->dst_roi[i] = 0;
        }
        av_log(NULL, AV_LOG_VERBOSE, "Disabling ROI, activations on current frame is %d, higher then cutoff %d.\n",
               total_activations_per, s->cutoff_per);
    }

    if (s->roi_map_val == ML_ROI_DQP || s->roi_map_val == ML_ROI_DQP_AND_RAW) {
        VpiSidedataHandle dst_dqp_sd = NULL;
        vpi_sidedata_alloc(s->sd_pool_mgr, VPI_SB_DQPMAP_HEVC_H264, s->dst_roi_size, RC_SIDE_EN, &dst_dqp_sd);

        unsigned char *dst_dqp = vpi_get_sidedata_vaddr(dst_dqp_sd);
        for (int i = 0; i < s->dst_roi_size; ++i) {
            dst_dqp[i] = -((max(0, (s->dst_roi[i]) - 128) * s->strength_val + 32) >> 6);
        }
        sb_scale_attribution(src_roi_sd, dst_dqp_sd, s->dst_roi_width, s->dst_roi_height);
        vpifrm_attach_sidedata(out_pic, dst_dqp_sd);
    }
    if (s->roi_map_val == ML_ROI_RAW || s->roi_map_val == ML_ROI_DQP_AND_RAW) {
        VpiSidedataHandle dst_roi_sd = NULL;
        vpi_sidedata_alloc(s->sd_pool_mgr, VPI_SB_ML_ROI, s->dst_roi_size, RC_SIDE_EN, &dst_roi_sd);
        memcpy(vpi_get_sidedata_vaddr(dst_roi_sd), s->dst_roi, s->dst_roi_size);
        sb_scale_attribution(src_roi_sd, dst_roi_sd, s->dst_roi_width, s->dst_roi_height);
        vpifrm_attach_sidedata(out_pic, dst_roi_sd);
    }

    for (int i = 1; i < s->nb_inputs; i++) {
        roi_pic = (VpiFrm *)in[i]->data[0];
        vpi_frame_unref(roi_pic);
    }

    return ff_filter_frame(outlink, out);
}


static int roi_scale_config_output(AVFilterLink *link)
{
    ROIScaleContext *s = NULL;
    AVFilterContext *ctx = NULL;
    FFFrameSyncIn *in = NULL;
    int ret;

    s = link->src->priv;
    s->fs.on_event = process_frame;
    s->fs.opaque   = s;
    ctx = link->src;

    s->src_roi_width = (ctx->inputs[1]->w + 15) >> 4;
    s->src_roi_height = (ctx->inputs[1]->h + 15) >> 4;
    s->src_roi_size = s->src_roi_width * s->src_roi_height;
    s->src_roi = (unsigned char *)malloc(sizeof(unsigned char) * s->src_roi_size);

    s->dst_roi_width = (ctx->inputs[0]->w + 15) >> 4;
    s->dst_roi_height = (ctx->inputs[0]->h + 15) >> 4;
    s->dst_roi_size = s->dst_roi_width * s->dst_roi_height;
    s->dst_roi = (unsigned char *)malloc(sizeof(unsigned char) * s->dst_roi_size);

    if (s->static_map_file) {
        FILE *fp = NULL;
        fp = fopen(s->static_map_file, "r");
        if (!fp) {
            av_log(ctx, AV_LOG_ERROR, "ROI file %s not found.\n", s->static_map_file);
            return AVERROR(EINVAL);
        }

        fseek(fp, 0L, SEEK_END);
        unsigned int static_roi_size = ftell(fp);
        if (s->src_roi_size != static_roi_size) {
            av_log(ctx, AV_LOG_ERROR, "Static ROI dimension does not match with the first input\n");
            return AVERROR(EINVAL);
        }

        s->static_roi_map = (unsigned char *)malloc(sizeof(unsigned char) * s->src_roi_size);
        rewind(fp);
        for (int i=0; i<static_roi_size; ++i) {
            char val;
            if (fscanf(fp, "%c", &val)) {
                s->static_roi_map[i] = val;
            }
        }
        fclose(fp);
    }

    ret = ff_framesync_init(&s->fs, ctx, s->nb_inputs);
    if (ret < 0)
        return ret;

    in = s->fs.in;
    for (int i = 0; i < s->nb_inputs; i++) {
        in[i].time_base = ctx->inputs[i]->time_base;
	in[i].sync   = 1;
	in[i].before = EXT_STOP;
	in[i].after  = EXT_INFINITY;
    }

    ret = ff_framesync_configure(&s->fs);
    link->time_base = s->fs.time_base;

    return ret;
}

static int config_output(AVFilterLink *link)
{
    int ret;
    ROIScaleContext *s = link->src->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, s->instance_name);
    ret = roi_scale_config_output(link);
    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ret;
}

static int roi_scale_activate(AVFilterContext *ctx)
{
    ROIScaleContext *s = ctx->priv;
    return ff_framesync_activate(&s->fs);
}

static int activate(AVFilterContext *ctx)
{
    int ret;
    ROIScaleContext *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, s->instance_name);
    ret = roi_scale_activate(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ret;
}


#define OFFSET(x) offsetof(ROIScaleContext, x)
#define FLAGS AV_OPT_FLAG_FILTERING_PARAM|AV_OPT_FLAG_VIDEO_PARAM

static const AVOption roi_scale_options[] = {
    { "nb_inputs", "set number of inputs", OFFSET(nb_inputs), AV_OPT_TYPE_INT, {.i64=2},  2, ROI_SCALE_MAX_INPUTS, .flags = FLAGS },
    { "static_map_file", "Provide ROI from file", OFFSET(static_map_file), AV_OPT_TYPE_STRING, {.str=NULL}, .flags = FLAGS},
    { "roi_map_type", "type of ROI map", OFFSET(roi_map_type), AV_OPT_TYPE_STRING, {.str = "dqp"},  0, 0, .flags = FLAGS },
    { "cutoff", "Activation threshold count per 100000, above which ROI is disabled (To configure 5.67%, use value 5670)", OFFSET(cutoff_per), AV_OPT_TYPE_INT, {.i64=5000}, 0, 100000, .flags = FLAGS},
    { "strength", "ROI strength : Low, Normal, High", OFFSET(strength), AV_OPT_TYPE_STRING, {.str = "NORMAL"}, 0, 0, .flags = FLAGS},
    { NULL }
};

AVFILTER_DEFINE_CLASS(roi_scale);

static const AVFilterPad roi_scale_outputs[] = {
    {
        .name         = "default",
        .type         = AVMEDIA_TYPE_VIDEO,
        .config_props = config_output,
    },
    { NULL }
};

const AVFilter ff_vf_roi_scale = {
    .name            = "roi_scale_ama",
    .description     = NULL_IF_CONFIG_SMALL("Scale the roi sidedata and move it on frame to be encoded."),
    .priv_size       = sizeof(ROIScaleContext),
    .priv_class      = &roi_scale_class,
    .init            = init,
    .uninit          = uninit,
    .activate        = activate,
    .outputs         = roi_scale_outputs,
    .nb_outputs      = 1,
    FILTER_QUERY_FUNC(query_formats),
    .flags           = AVFILTER_FLAG_SUPPORT_TIMELINE_INTERNAL,
};
