/*
* AMD side data debug filter
*
* This file is part of FFmpeg.
*
* FFmpeg is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* FFmpeg is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */


#include "vpi_types.h"
#include "vpi_api.h"

#include "avfilter.h"
#include "filters.h"
#include "libavutil/opt.h"
#include "libavutil/pixdesc.h"
#include "libavutil/hwcontext.h"

#include "log_ama.h"


struct SDDebugContext;

typedef struct SDDebugContext {
    const AVClass *class;
    char *tag_inst; //< string to identify unique instance
    int frame_cnt;
} SDDebugContext;


static int filter_frame(AVFilterLink *inlink, AVFrame *frame)
{
    AVFilterContext *ctx = inlink->dst;
    SDDebugContext *s = ctx->priv;
    AVHWFramesContext *hwframe_ctx;
    AVBufferRef *hw_frames_ctx;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_GENERAL, s->tag_inst);

    hw_frames_ctx = frame->hw_frames_ctx;
    if (hw_frames_ctx)
        hwframe_ctx = frame->hw_frames_ctx->data;
    if (hw_frames_ctx && hwframe_ctx && hwframe_ctx->format == AV_PIX_FMT_VPE) {
        VpiFrm *vpiframe = (VpiFrm*)frame->data[0];
        VpiSidedataHandle sb_info = NULL;

        while (vpiframe) {
            LOG_AMA_DEBUG(LOG_AMA_LAYER_FFMPEG, "Checking VPE sidedata for sd_debug_ama-%s, Frame number:%d, VpiFrm:%p, VpiFrm-ref:%d, AVFrame:%p, AVFrame-ref: %d\n", s->tag_inst, vpiframe->log_frm_num, vpiframe, vpi_frame_get_refcount(vpiframe), frame, av_buffer_get_ref_count(frame->buf[0]));

            for (int sb_type = 0; sb_type < VPI_SB_TYPE_MAX; sb_type++){
                vpifrm_get_first_sidedata_by_type((VpiFrm *)vpiframe, sb_type, &sb_info);
                while (sb_info){
                    LOG_AMA_DEBUG(LOG_AMA_LAYER_FFMPEG, "sd_debug_ama-%s, Frame number:%d, VpiFrm:%p, sidedata type: %d, sidedata ptr: %p, sidedata refcount: %d\n", s->tag_inst, vpiframe->log_frm_num, vpiframe, sb_type, vpi_get_sidedata_vaddr(sb_info),  vpi_sidedata_get_refcount(sb_info));
                    sb_info = vpifrm_find_next_sidedata((VpiFrm *)vpiframe, sb_info);
                }
            }
            vpiframe = vpiframe->next;
        }
    }
    else {
        LOG_AMA_DEBUG(LOG_AMA_LAYER_FFMPEG, "Checking AV sidedata for sd_debug_ama-%s, Frame number:%d, AVFrame:%p, AVFrame-ref: %d\n", s->tag_inst, s->frame_cnt, frame, av_buffer_get_ref_count(frame->buf[0]));
        for (int i = 0; i < frame->nb_side_data; i++) {
            if (frame->side_data[i]) {
                LOG_AMA_DEBUG(LOG_AMA_LAYER_FFMPEG, "sd_debug_ama-%s, Frame number:%d, AVFrame:%p, sidedata type: %d, sidedata ptr: %p, sidedata refcount: %d\n", s->tag_inst, s->frame_cnt, frame, frame->side_data[i]->type, frame->side_data[i],  av_buffer_get_ref_count(frame->side_data[i]->buf));
            }
        }
        s->frame_cnt++;
    }

    LOG_AMA_RESTORE_INSTANCE_INFO();

    return ff_filter_frame(inlink->dst->outputs[0], frame);
}


#define OFFSET(x) offsetof(SDDebugContext, x)
#define FLAGS (AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM)

static const AVOption sd_debug_options[] = {
    { "tag", "instance string",    OFFSET(tag_inst), AV_OPT_TYPE_STRING, { .str="SD_DEBUG" }, .flags = FLAGS},
    { NULL }
};


AVFILTER_DEFINE_CLASS(sd_debug);

static const AVFilterPad sd_debug_inputs[] = {
    {
        .name           = "default",
        .type           = AVMEDIA_TYPE_VIDEO,
        .filter_frame   = filter_frame,
    },
};

static const AVFilterPad sd_debug_outputs[] = {
    {
        .name = "default",
        .type = AVMEDIA_TYPE_VIDEO,
    },
};

const AVFilter ff_vf_sd_debug = {
    .name          = "sd_debug_ama",
    .description   = NULL_IF_CONFIG_SMALL("Debug the sidedata information."),
    .priv_size     = sizeof(SDDebugContext),
    .priv_class    = &sd_debug_class,
    .nb_inputs     = 1,
    .nb_outputs    = 1,
    .flags         = AVFILTER_FLAG_SUPPORT_TIMELINE_GENERIC,
    FILTER_INPUTS(sd_debug_inputs),
    FILTER_OUTPUTS(sd_debug_outputs),
};
