/*
 * AMA sideband data read filter
 * Copyright (C) 2023 VeriSilicon Holdings Co., Ltd.
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file
 * AMA sideband data read filter
 */
#include "vf_sd_common.h"

#define FLAGS AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM

static av_cold int init(AVFilterContext *ctx)
{
    AmaSdContext *sb = NULL;

    sb = ctx->priv;
    sb->sb_fp = fopen(sb->sb_file, "rb");
    if (!sb->sb_fp){
        av_log(ctx, AV_LOG_ERROR, "sdread_ama filter open sd file failed.\n");
        return AVERROR(EINVAL);
    }
    sb->frameId = -1;
    return 0;
}

static av_cold void uninit(AVFilterContext *ctx)
{
    AmaSdContext *sb = ctx->priv;
    if (sb)
        fclose(sb->sb_fp);
}

static int read_sidedata_file(AmaSdContext *sb, AVFrame *in, AVFrame *out){
    int ret;
    uint32_t data_size;
    enum AVFrameSideDataType sd_type;

    if(feof(sb->sb_fp))
        return 0;

    // first read
    if(sb->frameId == -1) {
        ret = fread(&sb->frameId, sizeof(uint32_t), 1, sb->sb_fp);
        if(ret != 1){
            if(feof(sb->sb_fp))
                return 0;
            else
                goto fail;
        }
    }

    av_log(NULL, AV_LOG_DEBUG, "frameId %d, coded_picture_number %d\n", sb->frameId, in->coded_picture_number);

    while(sb->frameId == in->coded_picture_number) {
        // read sidedata
        if(fread(&sd_type, sizeof(uint32_t), 1, sb->sb_fp) != 1)
            goto fail;
        if(fread(&data_size, sizeof(uint32_t), 1, sb->sb_fp) != 1)
            goto fail;
        av_log(NULL, AV_LOG_DEBUG, "sidedata type %d, data_size %d\n", sd_type, data_size);
        switch (sd_type)
        {
            case AV_FRAME_DATA_MASTERING_DISPLAY_METADATA: {
                AVMasteringDisplayMetadata *meta;
                meta = av_mastering_display_metadata_create_side_data(out);
                if(!meta)
                    goto fail;
                if (fread(meta, sizeof(uint8_t), data_size, sb->sb_fp) != data_size)
                    goto fail;
                break;
            }
            case AV_FRAME_DATA_CONTENT_LIGHT_LEVEL: {
                AVContentLightMetadata *meta;
                meta = av_content_light_metadata_create_side_data(out);
                if(!meta)
                    goto fail;
                if (fread(meta, sizeof(uint8_t), data_size, sb->sb_fp) != data_size)
                    goto fail;
                break;
            }
            case AV_FRAME_DATA_DYNAMIC_HDR_PLUS: {
                AVDynamicHDRPlus *hdr10plus;
                hdr10plus = av_dynamic_hdr_plus_create_side_data(out);
                if(!hdr10plus)
                    goto fail;
                if (fread(hdr10plus, sizeof(uint8_t), data_size, sb->sb_fp) != data_size)
                    goto fail;
                break;
            }
            case AV_FRAME_DATA_RAW_T35: {
                AVRawT35Data *raw_data;
                raw_data = av_raw_t35_data_create_side_data(out);
                if(!raw_data)
                    goto fail;
                if (fread(&raw_data->payload_byte_length, sizeof(uint32_t), 1, sb->sb_fp) != 1)
                    goto fail;
                raw_data->payload_byte_data = av_malloc(data_size);
                if(!raw_data->payload_byte_data)
                    goto fail;
                if (fread(raw_data->payload_byte_data, sizeof(uint8_t), data_size, sb->sb_fp) != data_size)
                    goto fail;
                break;
            }
            case AV_FRAME_DATA_VUI_PARAMS: {
                AVFrameSideData *side_data;
                side_data = av_frame_new_side_data(out, sd_type, data_size);
                if (!side_data){
                    av_log(sb, AV_LOG_ERROR, "sdread_ama filter failed to new sidedata.\n");
                    goto fail;
                }
                if (fread(side_data->data, sizeof(uint8_t), data_size, sb->sb_fp) != data_size)
                    goto fail;
                break;
            }
            case AV_FRAME_DATA_SB_ML_ROI:
            case AV_FRAME_DATA_SB_ML_CC:
            case AV_FRAME_DATA_SB_ML_CXC:
            case AV_FRAME_DATA_SB_ML_LA_STAT:
            case AV_FRAME_DATA_SB_ML_BOUNDING_BOX:
            case AV_FRAME_DATA_SB_ML_CLASSIFICATION:
            case AV_FRAME_DATA_SB_ML_YDATA:
            case AV_FRAME_DATA_SB_ML_TENSOR: {
                AVFrameSideData *side_data;
                side_data = av_frame_new_side_data(out, sd_type, data_size);
                if (!side_data){
                    av_log(sb, AV_LOG_ERROR, "sdread_ama filter failed to new sidedata.\n");
                    goto fail;
                }
                if (fread(&sb->ml_header, ML_HEAD_SIZE, 1, sb->sb_fp) != 1)
                    goto fail;
                av_dict_set_int(&side_data->metadata, "width",   sb->ml_header.width,    0);
                av_dict_set_int(&side_data->metadata, "height",  sb->ml_header.height,   0);
                av_dict_set_int(&side_data->metadata, "channel", sb->ml_header.channel,  0);
                av_dict_set_int(&side_data->metadata, "batch",   sb->ml_header.batch,    0);
                av_dict_set_int(&side_data->metadata, "layout",  sb->ml_header.layout,   0);
                if (fread(side_data->data, sizeof(uint8_t), data_size, sb->sb_fp) != data_size)
                    goto fail;
                break;
            }
            default:
                break;
        }
        ret = fread(&sb->frameId, sizeof(uint32_t), 1, sb->sb_fp);
        if(ret != 1){
            if(feof(sb->sb_fp))
                return 0;
            else
                goto fail;
        }
    }

    return 0;
fail:
    return AVERROR(AVERROR_EXTERNAL);
}

static int filter_frame(AVFilterLink *inlink, AVFrame *in)
{
    AVFilterLink *outlink = inlink->dst->outputs[0];
    AVFrame *out = ff_get_video_buffer(outlink, in->width, in->height);
    AmaSdContext *sb = (AmaSdContext *)inlink->dst->priv;
    int ret;

    if (!out) {
        ret = AVERROR(ENOMEM);
        goto fail;
    }

    ret = av_frame_copy_props(out, in);
    if (ret < 0)
        goto fail;
    ret = av_frame_copy(out, in);
    if (ret < 0)
        goto fail;

    ret = read_sidedata_file(sb, in, out);
    if (ret < 0)
        goto fail;

    av_frame_free(&in);
    return ff_filter_frame(outlink, out);
fail:
    av_frame_free(&in);
    av_frame_free(&out);
    return ret;
}

static const AVOption sd_read_options[] = {
    { "file",   "sidedata file path", OFFSET(sb_file), AV_OPT_TYPE_STRING, { .str = NULL },  .flags = FLAGS },
    { NULL }
};

AVFILTER_DEFINE_CLASS(sd_read);

static const AVFilterPad avfilter_vf_sd_read_inputs[] = {
    {
        .name         = "default",
        .type         = AVMEDIA_TYPE_VIDEO,
        .filter_frame = filter_frame,
    },
};

static const AVFilterPad avfilter_vf_sd_read_outputs[] = {
    {
        .name = "default",
        .type = AVMEDIA_TYPE_VIDEO,
    },
};

const AVFilter ff_vf_sd_read = {
    .name        = "sdread_ama",
    .description = NULL_IF_CONFIG_SMALL("Read sidedata from file to the output."),
    .flags       = AVFILTER_FLAG_METADATA_ONLY,
    FILTER_INPUTS(avfilter_vf_sd_read_inputs),
    FILTER_OUTPUTS(avfilter_vf_sd_read_outputs),
    .priv_size   = sizeof(AmaSdContext),
    .priv_class  = &sd_read_class,
    .init        = init,
    .uninit      = uninit,
};